# -*- coding: utf-8 -*-

"""Implementation of HttpTagQueryResultCollection."""

from typing import Any, Dict, List, Optional, Tuple

from nisystemlink.clients import tag as tbase
from nisystemlink.clients.core._internal._http_client import HttpClient, HttpResponse
from typing_extensions import final


@final
class HttpTagQueryResultCollection(tbase.TagQueryResultCollection):
    def __init_subclass__(cls) -> None:
        raise TypeError(
            "type 'HttpTagQueryResultCollection' is not an acceptable base type"
        )

    def __init__(
        self,
        client: HttpClient,
        paths: Optional[str],
        keywords: Optional[str],
        properties: Optional[str],
        skip: int,
        take: Optional[int],
        tag_query_result: Dict[str, Any],
        http_response: HttpResponse,
    ) -> None:
        first_page, total_count = self.__handle_query_response(
            tag_query_result, http_response
        )
        super().__init__(first_page, total_count, skip)

        api = client.at_uri("/nitag/v2")
        base_params = {
            "path": paths,
            "keywords": keywords,
            "properties": properties,
            "skip": "0",
            "take": str(take) if take is not None else None,
        }
        for k, v in list(base_params.items()):
            if v is None:
                del base_params[k]

        def query(s: int) -> Tuple[Dict[str, Any], HttpResponse]:
            params = dict(base_params)
            params["skip"] = str(s)
            return api.get("/tags", params=params)

        self._query = query

    def _query_page(self, skip: int) -> List[tbase.TagData]:
        page, self._total_count = self.__handle_query_response(*self._query(skip))
        return page

    def __handle_query_response(
        self, response: Dict[str, Any], http_response: HttpResponse
    ) -> Tuple[List[tbase.TagData], int]:
        if response.get("totalCount") is None:
            raise tbase.TagManager.invalid_response(http_response)

        tags = []
        for t in response["tags"]:
            tags.append(
                tbase.TagData(
                    t["path"],
                    tbase.DataType.from_api_name(t["type"]) if t["type"] else None,
                    t.get("keywords"),
                    t.get("properties"),
                )
            )
            if t.get("collectAggregates"):
                tags[-1].collect_aggregates = True
        return tags, response["totalCount"]



# -*- coding: utf-8 -*-

"""Implementation of SystemTimeStamper."""

import datetime
import threading

from nisystemlink.clients.tag._core._itime_stamper import ITimeStamper
from typing_extensions import final


@final
class SystemTimeStamper(ITimeStamper):
    """A :class:`ITimeStamper` that uses the system clock."""

    _increment = datetime.timedelta(microseconds=1)

    def __init_subclass__(cls) -> None:
        raise TypeError("type 'SystemTimeStamper' is not an acceptable base type")

    def __init__(self) -> None:
        self._lock = threading.Lock()
        self._last = datetime.datetime.now(datetime.timezone.utc)

    @property
    def timestamp(self) -> datetime.datetime:  # noqa: D401
        """A unique UTC ``datetime.datetime`` that is close to the current date and time."""
        timestamp = datetime.datetime.now(datetime.timezone.utc)

        with self._lock:
            if timestamp > self._last:
                self._last = timestamp
            else:
                self._last += self._increment
            return self._last



from typing import Dict, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._column_type import ColumnType
from ._data_type import DataType


class Column(JsonModel):
    """Defines a single column in a table."""

    name: str
    """The column name, which must be unique across all columns in the table."""

    data_type: DataType
    """The data type of the column."""

    column_type: ColumnType = ColumnType.Normal
    """The column type. Defaults to ColumnType.Normal."""

    properties: Optional[Dict[str, str]] = None
    """User-defined properties associated with the column."""



# -*- coding: utf-8 -*-

"""Implementation of PathConstants."""

import os
import pathlib
import typing
from typing import Optional

from nisystemlink.clients.core._internal._classproperty_support import (
    ClasspropertySupport,
)
from typing_extensions import final


@final
class PathConstants(metaclass=ClasspropertySupport):
    """Provides file and directory paths for the SystemLink client APIs."""

    COMPANY_NAME = "National Instruments"

    PRODUCT_NAME = "Skyline"

    _application_data_directory = None  # type: Optional[pathlib.Path]

    def __init_subclass__(cls) -> None:
        raise TypeError("type 'PathConstants' is not an acceptable base type")

    def __init__(self) -> None:
        raise TypeError("Can't instantiate static class 'PathConstants'")

    @ClasspropertySupport.classproperty
    def application_data_directory(self) -> pathlib.Path:  # noqa: D401
        """Get the platform-specific path to the common NI Application Data directory.

        Returns:
            The platform-specific path to the common NI Application Data directory.
        """
        if self._application_data_directory is None:
            if os.name == "nt":
                PathConstants._application_data_directory = (
                    self._windows_application_data_directory()
                )
            else:
                PathConstants._application_data_directory = pathlib.Path(
                    "/etc/natinst/niskyline"
                )
        return typing.cast(pathlib.Path, self._application_data_directory)

    @classmethod
    def _windows_application_data_directory(cls) -> pathlib.Path:
        """Get the NI Application Data directory on Windows.

        Returns:
            The NI Application Data directory on Windows.

        Raises:
            RuntimeError: if this method is called on a non-Windows system.
            RuntimeError: if the ProgramData folder cannot be found.
        """
        if os.name != "nt":
            raise RuntimeError("This function is for Windows only")

        try:
            from . import _winpaths

            programdata_dir = pathlib.Path(
                _winpaths.get_path(_winpaths.FOLDERID.ProgramData)
            )
        except Exception:
            programdata_dir = pathlib.Path("C:/ProgramData")
        if not programdata_dir.exists():
            raise RuntimeError("Cannot find ProgramData folder")

        return programdata_dir / cls.COMPANY_NAME / cls.PRODUCT_NAME



# -*- coding: utf-8 -*-

"""Implementation of TagUpdateFields."""

import sys

if sys.version_info < (3, 6):
    import aenum as enum  # type: ignore
else:
    import enum


class TagUpdateFields(enum.IntFlag):
    """Represents the various :class:`TagData` fields that may be included in a :class:`TagDataUpdate`.

    Fields that aren't included are left unmodified when updates are sent to the server.
    """

    KEYWORDS = 1
    """Specify that entries in the :attr:`TagData.keywords` that are missing from the
    tag's metadata on the server will be added.
    """

    PROPERTIES = 2
    """Specify that entries in the :attr:`TagData.properties` that are missing from or
    different in the tag's metadata on the server will be added or replaced.
    """

    COLLECT_AGGREGATES = 4
    """Specify that the tag's :attr:`TagData.collect_aggregates` setting will be modified on the server."""

    RETENTION = 8
    """Specify that the tag's :attr:`TagData.retention_type`,
    :attr:`TagData.retention_count`, and :attr:`TagData.retention_days` settings will be
    modified on the server.
    """

    ALL = 15
    """Specify that all fields in the :class:`TagData` should be included in the update.

    Keywords and properties that already exist on the server will not be removed, even
    if they are missing from the update.
    """



from enum import Enum
from typing import List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._column_filter import ColumnFilter
from ._column_order_by import ColumnOrderBy


class ExportFormat(str, Enum):
    """The format of the exported data."""

    CSV = "CSV"
    """Comma-separated values."""


class ExportTableDataRequest(JsonModel):
    """Specifies the parameters for a data export with ordering and filtering."""

    columns: Optional[List[str]] = None
    """The names of columns to include in the export. The export will
    include the columns in the same order specified in this parameter. All
    columns are included in the order specified at table creation if this
    property is excluded."""

    order_by: Optional[List[ColumnOrderBy]] = None
    """A list of columns to order the results by. Multiple columns may be
    specified to order rows that have the same value for prior columns. The
    columns used for sorting do not need to be included in the columns list, in
    which case they are not included in the export."""

    filters: Optional[List[ColumnFilter]] = None
    """A list of columns to filter by. Only rows whose columns contain values
    matching all of the specified filters are returned. The columns used for
    filtering do not need to be included in the columns list, in which case
    they are not included in the export."""

    response_format: ExportFormat
    """The format of the exported data. The only response format
    currently supported is ``CSV``."""



# -*- coding: utf-8 -*-

"""Implementation of TagData."""

from typing import Any, Dict, Iterable, List, Optional

from nisystemlink.clients import tag as tbase
from typing_extensions import final


@final
class TagData:
    """Contains the metadata for a SystemLink tag."""

    _RETENTION_TYPE_PROP = "nitagRetention"

    _RETENTION_TYPE_NONE = "NONE"

    _RETENTION_TYPE_DURATION = "DURATION"

    _RETENTION_TYPE_COUNT = "COUNT"

    _RETENTION_TYPE_PERMANENT = "PERMANENT"

    _HISTORY_TTL_DAYS_PROP = "nitagHistoryTTLDays"

    _MAX_HISTORY_COUNT_PROP = "nitagMaxHistoryCount"

    def __init_subclass__(cls) -> None:
        raise TypeError("type 'TagData' is not an acceptable base type")

    def __init__(
        self,
        path: str,
        data_type: Optional[tbase.DataType] = None,
        keywords: Optional[Iterable[str]] = None,
        properties: Optional[Dict[str, str]] = None,
    ) -> None:
        """Initialize an instance.

        Args:
            path: The tag's path, which uses a dot-separated hierarchy to uniquely
                identify the tag on the server.
            data_type: The data type for the tag's values.
            keywords: The tag's keywords.
            properties: The tag's properties.
        """
        self._path = path
        self._data_type = tbase.DataType.UNKNOWN if data_type is None else data_type
        self._keywords = list(keywords) if keywords else []
        self._properties = {}  # type: Dict[str, str]
        self._collect_aggregates = False
        self._retention_type = tbase.RetentionType.NONE
        self._retention_count = None  # type: Optional[int]
        self._retention_days = None  # type: Optional[int]
        if properties:
            self.replace_properties(properties)

    @classmethod
    def from_json_dict(cls, data: Dict[str, Any]) -> "TagData":
        data_type_str = data.get("type") or "UNKNOWN"
        data_type = tbase.DataType.from_api_name(data_type_str)
        tag = cls(data["path"], data_type, data.get("keywords"), data.get("properties"))
        if data.get("collectAggregates"):
            tag.collect_aggregates = True
        return tag

    def to_json_dict(self) -> Dict[str, Any]:
        self.validate_path()
        if self.data_type == tbase.DataType.UNKNOWN:
            raise ValueError("Invalid tag data type")

        data = {}  # type: Dict[str, Any]
        data["path"] = self._path
        data["type"] = self._data_type.api_name
        data["collectAggregates"] = self.collect_aggregates

        if self._keywords:
            data["keywords"] = self._keywords

        data["properties"] = dict(self._properties) if self._properties else {}
        self._copy_retention_properties(data["properties"])
        if not data["properties"]:
            del data["properties"]

        return data

    @property
    def collect_aggregates(self) -> bool:  # noqa: D401
        """Whether the server should keep aggregate information for the tag.

        The information collected depends on the tag's :attr:`data_type`.
        """
        return self._collect_aggregates

    @collect_aggregates.setter
    def collect_aggregates(self, value: bool) -> None:
        self._collect_aggregates = value

    @property
    def data_type(self) -> tbase.DataType:  # noqa: D401
        """The data type for the tag's values.

        Changing the data type of an existing tag requires deleting the tag and creating
        a new one of a different data type.
        """
        return self._data_type

    @data_type.setter
    def data_type(self, value: tbase.DataType) -> None:
        self._data_type = value

    @property
    def keywords(self) -> List[str]:  # noqa: D401
        """The list of keywords associated with the tag."""
        return self._keywords

    @property
    def path(self) -> str:  # noqa: D401
        """The tag's path, which uses a dot-separated hierarchy to uniquely identify the tag on the server."""
        return self._path

    @property
    def properties(self) -> Dict[str, str]:  # noqa: D401
        """The properties associated with the tag."""
        return self._properties

    @property
    def retention_type(self) -> tbase.RetentionType:  # noqa: D401
        """How the tag's historical values are retained by the tag historian, if available.

        The :attr:`retention_count` and :attr:`retention_days` properties can further
        customize when values are removed from the historian.

        The tag historian is an optional component for SystemLink Server installations,
        and is not available in SystemLink Cloud. The tag's historical values are not
        retained when the tag historian is not available, regardless of the retention
        type.
        """
        return self._retention_type

    @retention_type.setter
    def retention_type(self, value: tbase.RetentionType) -> None:
        self._retention_type = value

    @property
    def retention_count(self) -> Optional[int]:  # noqa: D401
        """The number of historical values to retain when :attr:`retention_type` is
        :attr:`RetentionType.COUNT`, or None to use the server-specified default of
        10000.
        """
        return self._retention_count

    @retention_count.setter
    def retention_count(self, value: Optional[int]) -> None:
        self._retention_count = value

    @property
    def retention_days(self) -> Optional[int]:  # noqa: D401
        """The number of days to keep a tag's historical values when
        :attr:`retention_type` is :attr:`RetentionType.DURATION`, or None to use the
        server-specified default of 30 days.
        """
        return self._retention_days

    @retention_days.setter
    def retention_days(self, value: Optional[int]) -> None:
        self._retention_days = value

    def replace_keywords(self, keywords: Iterable[str]) -> None:
        """Replace all of the tag's :attr:`keywords` with those in ``keywords``.

        Args:
            keywords: The tag's new keywords, or None to clear all keywords.
        """
        self._keywords[:] = keywords

    def replace_properties(self, properties: Dict[str, str]) -> None:
        """Replace all of the tag's :attr:`properties` with those in ``properties``.

        Args:
            properties: The tag's new properties, or None to clear all properties.
        """
        self._properties.clear()

        if properties is None:
            return

        for key, value in properties.items():
            # Check for special properties to parse.
            if key == self._RETENTION_TYPE_PROP:
                if value == self._RETENTION_TYPE_COUNT:
                    self._retention_type = tbase.RetentionType.COUNT
                elif value == self._RETENTION_TYPE_DURATION:
                    self._retention_type = tbase.RetentionType.DURATION
                elif value == self._RETENTION_TYPE_NONE:
                    self._retention_type = tbase.RetentionType.NONE
                elif value == self._RETENTION_TYPE_PERMANENT:
                    self._retention_type = tbase.RetentionType.PERMANENT
                else:
                    self._retention_type = tbase.RetentionType.INVALID
            elif key == self._HISTORY_TTL_DAYS_PROP:
                try:
                    self._retention_days = int(value)
                except ValueError:
                    self._retention_days = None
            elif key == self._MAX_HISTORY_COUNT_PROP:
                try:
                    self._retention_count = int(value)
                except ValueError:
                    self._retention_count = None
            else:
                # Not a special property. Preserve it in the dictionary.
                self._properties[key] = value

    def _copy_retention_properties(self, destination: Dict[str, str]) -> None:
        """Copy the tag's retention settings into ``destination``.

        Clients do not typically call this method directly.

        Args:
            destination: The dictionary to copy into. Existing properties with names in
                common with the retention properties will be overwritten.

        Raises:
            ValueError: if ``destination`` is None.
        """
        if destination is None:
            raise ValueError("destination cannot be None")

        if self._retention_type == tbase.RetentionType.NONE:
            destination[self._RETENTION_TYPE_PROP] = self._RETENTION_TYPE_NONE
        elif self._retention_type == tbase.RetentionType.DURATION:
            destination[self._RETENTION_TYPE_PROP] = self._RETENTION_TYPE_DURATION
        elif self._retention_type == tbase.RetentionType.COUNT:
            destination[self._RETENTION_TYPE_PROP] = self._RETENTION_TYPE_COUNT
        elif self._retention_type == tbase.RetentionType.PERMANENT:
            destination[self._RETENTION_TYPE_PROP] = self._RETENTION_TYPE_PERMANENT

        if self._retention_count is not None:
            destination[self._MAX_HISTORY_COUNT_PROP] = str(self._retention_count)

        if self._retention_days is not None:
            destination[self._HISTORY_TTL_DAYS_PROP] = str(self._retention_days)

    def clear_retention(self) -> None:
        """Clear all retention settings, setting it to use a
        :attr:`TagData.retention_type` of :attr:`RetentionType.NONE`.

        Args:
            tag: The tag whose retention will be cleared.
        """
        self.retention_type = tbase.RetentionType.NONE
        self.retention_count = None
        self.retention_days = None

    def set_retention_count(self, count: int) -> None:
        """Set the number of historical values to retain.

        Args:
            count: The number of historical values to retain.
        """
        self.retention_type = tbase.RetentionType.COUNT
        self.retention_count = count
        self.retention_days = None

    def set_retention_days(self, days: int) -> None:
        """Set the historical values to be retained for the specified number of days.

        Args:
            days: The number of days a historical value will be kept.
        """
        self.retention_type = tbase.RetentionType.DURATION
        self.retention_count = None
        self.retention_days = days

    def validate_type(self, required_type: tbase.DataType) -> None:
        """Validate that the tag's data type matches ``required_type``.

        Clients do not typically call this method directly.

        Args:
            required_type: The data type required by the API.

        Raises:
            ValueError: if this is not a tag of the required type with a valid path.
            ValueError: if ``required_type`` is :attr:`DataType.UNKNOWN`.
        """
        if required_type == tbase.DataType.UNKNOWN:
            raise ValueError("required_type is not a valid data type")

        if self.data_type != required_type:
            raise ValueError(
                "Tag must be a {} type but is a {} tag".format(
                    required_type.name, self.data_type.name
                )
            )

    def validate_path(self) -> str:
        """Validate the path as an input and returns it.

        Clients do not typically call this method directly.

        Returns:
            The validated path.

        Raises:
            ValueError: if the tag's path is invalid.
        """
        if self.path is None:
            raise ValueError("Tag path cannot be None")

        return tbase.TagPathUtilities.validate(self.path)



# -*- coding: utf-8 -*-

"""Implementation of DataType."""

import enum


class DataType(enum.Enum):
    """Represents the different data types for a SystemLink tag."""

    UNKNOWN = 0
    """An unknown or invalid data type.

    Not a valid input to API calls, but used to represent tags whose data type isn't
    recognized.
    """

    DOUBLE = 1
    """A 64-bit floating-point tag following the IEEE standard.

    Double tags support collecting aggregate values for the min, max, mean, and count.
    """

    INT32 = 2
    """A 32-bit signed integer tag.

    Int32 tags support collecting aggregate values for the min, max, mean, and count.
    The mean is represented as a double.
    """

    STRING = 3
    """A string tag for arbitrary values.

    String tags support collecting aggregate values for the count.
    """

    BOOLEAN = 4
    """A boolean tag.

    Bool tags support collecting aggregate values for the count.
    """

    UINT64 = 5
    """A 64-bit unsigned integer tag.

    UInt64 tags support collecting aggregate values for the min, max, mean, and count.
    The mean is represented as a double value and will truncate large values.
    """

    DATE_TIME = 6
    """A date and time tag that stores values in UTC ISO 8601 format.

    DateTime tags support collecting aggregate values for the count.
    """

    @property
    def api_name(self) -> str:
        """Web API name of the enum value."""
        return _API_NAME[self]

    @classmethod
    def from_api_name(cls, name: str) -> "DataType":
        return cls(_FROM_API_NAME.get(name, cls.UNKNOWN))


_API_NAME = {
    DataType.UNKNOWN: "UNKNOWN",
    DataType.DOUBLE: "DOUBLE",
    DataType.INT32: "INT",
    DataType.STRING: "STRING",
    DataType.BOOLEAN: "BOOLEAN",
    DataType.UINT64: "U_INT64",
    DataType.DATE_TIME: "DATE_TIME",
}

_FROM_API_NAME = {
    "DOUBLE": DataType.DOUBLE,
    "INT": DataType.INT32,
    "STRING": DataType.STRING,
    "BOOLEAN": DataType.BOOLEAN,
    "U_INT64": DataType.UINT64,
    "DATE_TIME": DataType.DATE_TIME,
}



from typing import List

from nisystemlink.clients.core import ApiError
from nisystemlink.clients.core._uplink._json_model import JsonModel


class DeleteTablesPartialSuccess(JsonModel):
    """The result of deleting multiple tables when one or more tables could not be deleted."""

    deleted_table_ids: List[str]
    """The IDs of the tables that were successfully deleted."""

    failed_table_ids: List[str]
    """The IDs of the tables that could not be deleted."""

    error: ApiError
    """The error that occurred when deleting the tables."""



# -*- coding: utf-8 -*-

"""Implementation of HttpConfigurationManager."""

import json
import pathlib
import typing
from typing import Dict, Optional

from nisystemlink.clients import core
from nisystemlink.clients.core._internal._http_configuration_file import (
    HttpConfigurationFile,
)
from nisystemlink.clients.core._internal._path_constants import PathConstants


class HttpConfigurationManager:
    """Factory for :class:`HttpConfiguration` objects."""

    HTTP_MASTER_CONFIGURATION_ID = "SYSTEMLINK_MASTER"
    """The default ID of the SystemLink Server's configuration on systems registered using SystemLink Client."""

    HTTP_LOCALHOST_CONFIGURATION_ID = "SYSTEMLINK_LOCALHOST"
    """The default ID of the SystemLink Server's configuration on the SystemLink Server itself."""

    _configs = None

    @classmethod
    def get_configuration(
        cls, id: Optional[str] = None, enable_fallbacks: Optional[bool] = True
    ) -> core.HttpConfiguration:
        """Get the requested or default configuration.

        Args:
            id: The ID of the configuration to find.
            enable_fallbacks: Whether or not to fallback to other known configurations,
                if ``id`` is unavailable.

        Returns:
            The configuration.

        Raises:
            ValueError: if ``id`` is None and ``enable_fallbacks`` is False.
            ApiException: if the specified (or default) configuration is not found.
        """
        if id is not None:
            id = id.upper()
        else:
            if not enable_fallbacks:
                raise ValueError("id cannot be None if enable_fallbacks is False")
            fallback_configuration = cls._fallback()
            if fallback_configuration is None:
                raise core.ApiException("No SystemLink configurations available")
            return fallback_configuration

        if cls._configs is None:
            cls._configs = cls._read_configurations()
        config = cls._configs.get(id)
        if config is not None:
            return config
        if enable_fallbacks:
            fallback_configuration = cls._fallback()
            if fallback_configuration is None:
                raise core.ApiException(
                    "Configuration with ID {!r} was not found, and no ".format(id)
                    + "fallback configurations were available."
                )
            return fallback_configuration
        raise core.ApiException("Configuration with ID {!r} was not found.".format(id))

    @classmethod
    def _fallback(cls) -> Optional[core.HttpConfiguration]:
        """Attempt to acquire fallback HTTP configurations.

        Returns:
            The best available fallback configuration, or None if no such
            configurations are available.
        """
        if cls._configs is None:
            cls._configs = cls._read_configurations()
        master_config = cls._configs.get(cls.HTTP_MASTER_CONFIGURATION_ID)
        if master_config is not None:
            return master_config
        localhost_config = cls._configs.get(cls.HTTP_LOCALHOST_CONFIGURATION_ID)
        if localhost_config is not None:
            return localhost_config
        return None

    @classmethod
    def _read_configurations(cls) -> Dict[str, core.HttpConfiguration]:
        """Discover and loads the HTTP configuration files.

        Returns:
            A dictionary mapping each loaded configuration ID to its corresponding
            :class:`HttpConfiguration`.

        Raises:
            ApiException: if multiple configurations with the same ID are simultaneously
                present.
            ApiException: if an OS or permission error prevents reading the directory
                that contains HTTP configurations.
        """
        configurations = {}  # type: Dict[str, core.HttpConfiguration]
        path = cls._http_configurations_directory()
        if not path.exists():
            return configurations
        try:
            json_files = path.glob("*.json")
        except PermissionError as e:
            raise core.ApiException(
                "Not authorized to access HTTP configurations directory: " + str(e)
            )
        except OSError as e:
            raise core.ApiException(
                "Error while accessing HTTP configurations directory: " + str(e)
            )

        for json_file in json_files:
            try:
                config_file = cls._read_configuration_file(json_file)
                if config_file is None or config_file.id is None:
                    continue
                if config_file.id in configurations:
                    raise core.ApiException("Duplicate configuration IDs detected.")
                if not config_file.uri:
                    continue

                cert_path = None  # type: Optional[pathlib.Path]
                if config_file.cert_path:
                    cert_path = typing.cast(
                        pathlib.Path,
                        PathConstants.application_data_directory
                        / "Certificates"
                        / config_file.cert_path,
                    )
                    if not cert_path.exists():
                        cert_path = None
                configurations[config_file.id] = core.HttpConfiguration(
                    config_file.uri, config_file.api_key, cert_path=cert_path
                )
            except PermissionError:
                pass
            except OSError:
                # The individual file is inaccessible or badly formatted, so just skip it
                pass
            except json.JSONDecodeError:
                pass
            except ValueError:
                pass
        return configurations

    @classmethod
    def _read_configuration_file(
        cls, path: pathlib.Path
    ) -> Optional[HttpConfigurationFile]:
        """Parse a single SystemLink HTTP configuration file.

        Args:
            path: The path of the file.

        Returns:
            The parsed file, or None if there is no valid configuration present at the
            given path.

        Raises:
            OSError: if an error occurs while accessing the file.
            PermissionError: if the current application does not have permission to
                access the configuration file.
            json.decoder.JSONDecodeError: if the file does not contain valid JSON.
        """
        if not path.exists():
            raise OSError("HTTP configuration file was not found: " + str(path))
        with path.open() as f:
            config_dict = json.load(f)
        config = HttpConfigurationFile.from_json_dict(config_dict)
        if not config.id:
            return None

        config.id = config.id.upper()
        return config

    @classmethod
    def _http_configurations_directory(cls) -> pathlib.Path:
        """Get the platform-specific path to the HTTP Configurations directory.

        Returns:
            The path of the HTTP Configurations directory.
        """
        return PathConstants.application_data_directory / "HttpConfigurations"



# -*- coding: utf-8 -*-

"""Implementation of HttpClient."""

import json.decoder
import sys
import threading
import typing
import urllib.parse
from typing import Any, Awaitable, Dict, Iterable, Optional, Tuple, Union

from nisystemlink.clients import core

if sys.version_info >= (3, 6):
    from httpx import AsyncClient, Client, Response as HttpResponse
else:
    from requests import Session as Client, Response as HttpResponse

    AsyncClient = None  # type: Any


class HttpClient:
    """Base client for HTTP connections."""

    def __init__(self, configuration: core.HttpConfiguration) -> None:
        self._server = configuration.server_uri.rstrip("/")

        self._kwargs = {}  # type: Dict[str, Any]
        self._kwargs["headers"] = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        if configuration.api_keys is not None:
            self._kwargs["headers"].update(configuration.api_keys)
        elif configuration.username is not None and configuration.password is not None:
            self._kwargs["auth"] = (configuration.username, configuration.password)
        if configuration.cert_path:
            self._kwargs["verify"] = str(configuration.cert_path)

        # Keep a client per thread
        # - https://toolbelt.readthedocs.io/en/latest/threading.html
        # - "there are still a couple corner cases where it isn't perfectly threadsafe"
        self._clients = {}  # type: Dict[int, Client]
        self._aclients = {}  # type: Dict[int, AsyncClient]

    def at_uri(self, uri: str) -> "_HttpClientAtUri":
        """Get a client interface for which all queries are relative to ``uri``."""
        return _HttpClientAtUri(self, self._server + uri)

    @property
    def _client(self) -> Client:
        thread_id = threading.get_ident()
        if thread_id not in self._clients:
            if sys.version_info >= (3, 6):
                client = Client(**self._kwargs)
            else:
                client = Client()
                for k, v in self._kwargs.items():
                    setattr(client, k, v)
            self._clients[thread_id] = client
        return self._clients[thread_id]

    @property
    def _async_client(self) -> AsyncClient:
        thread_id = threading.get_ident()
        if thread_id not in self._clients:
            if sys.version_info < (3, 6):
                raise RuntimeError("async support is only available for python 3.6+")
            self._aclients[thread_id] = AsyncClient(**self._kwargs)
        return self._aclients[thread_id]


class _HttpClientAtUri:
    """Interface to HttpClient for while all queries are relative to a given uri."""

    def __init__(self, client: HttpClient, uri: str) -> None:
        self._client = client
        self._base_uri = uri

    @property
    def as_async(self) -> "_AsyncHttpClientAtUri":
        """An async version of the client."""
        if sys.version_info < (3, 6):
            raise RuntimeError("async support is only available for python 3.6+")
        return _AsyncHttpClientAtUri(self._client, self._base_uri)

    def _request(
        self,
        method: str,
        uri: str,
        *,
        params: Optional[Dict[str, Optional[str]]] = None,
        data: Optional[Union[Dict[str, Any], Iterable[Any]]] = None
    ) -> Tuple[Any, HttpResponse]:
        client = self._client._client
        uri, params2 = _expand_uri_params(uri, params)
        response = client.request(method, uri, json=data, params=params2)
        return _handle_response(response, method, uri), response

    def get(
        self, uri: str, *, params: Optional[Dict[str, Optional[str]]] = None
    ) -> Tuple[Any, HttpResponse]:
        """Perform a GET request."""
        return self._request("GET", self._base_uri + uri, params=params)

    def head(
        self, uri: str, *, params: Optional[Dict[str, Optional[str]]] = None
    ) -> Tuple[Any, HttpResponse]:
        """Perform a HEAD request."""
        return self._request("HEAD", self._base_uri + uri, params=params)

    def delete(
        self, uri: str, *, params: Optional[Dict[str, Optional[str]]] = None
    ) -> Tuple[Any, HttpResponse]:
        """Perform a DELETE request."""
        return self._request("DELETE", self._base_uri + uri, params=params)

    def post(
        self,
        uri: str,
        *,
        params: Optional[Dict[str, Optional[str]]] = None,
        data: Optional[Union[Dict[str, Any], Iterable[Any]]] = None
    ) -> Tuple[Any, HttpResponse]:
        """Perform a POST request."""
        return self._request("POST", self._base_uri + uri, params=params, data=data)

    def put(
        self,
        uri: str,
        *,
        params: Optional[Dict[str, Optional[str]]] = None,
        data: Optional[Union[Dict[str, Any], Iterable[Any]]] = None
    ) -> Tuple[Any, HttpResponse]:
        """Perform a PUT request."""
        return self._request("PUT", self._base_uri + uri, params=params, data=data)

    def patch(
        self,
        uri: str,
        *,
        params: Optional[Dict[str, Optional[str]]] = None,
        data: Optional[Union[Dict[str, Any], Iterable[Any]]] = None
    ) -> Tuple[Any, HttpResponse]:
        """Perform a PATCH request."""
        return self._request("PATCH", self._base_uri + uri, params=params, data=data)

    @property
    def base_uri(self) -> str:
        return self._base_uri


class _AsyncHttpClientAtUri:
    """Interface to HttpClient for while all queries are relative to a given uri."""

    def __init__(self, client: HttpClient, uri: str) -> None:
        self._client = client
        self._base_uri = uri

    async def _request(
        self,
        method: str,
        uri: str,
        *,
        params: Optional[Dict[str, Optional[str]]] = None,
        data: Optional[Union[Dict[str, Any], Iterable[Any]]] = None
    ) -> Tuple[Any, HttpResponse]:
        client = self._client._async_client
        uri, params2 = _expand_uri_params(uri, params)
        response = await client.request(method, uri, json=data, params=params2)
        return _handle_response(response, method, uri), response

    def get(
        self, uri: str, *, params: Optional[Dict[str, Optional[str]]] = None
    ) -> Awaitable[Tuple[Any, HttpResponse]]:
        """Perform a GET request."""
        return self._request("GET", self._base_uri + uri, params=params)

    def head(
        self, uri: str, *, params: Optional[Dict[str, Optional[str]]] = None
    ) -> Awaitable[Tuple[Any, HttpResponse]]:
        """Perform a HEAD request."""
        return self._request("HEAD", self._base_uri + uri, params=params)

    def delete(
        self, uri: str, *, params: Optional[Dict[str, Optional[str]]] = None
    ) -> Awaitable[Tuple[Any, HttpResponse]]:
        """Perform a DELETE request."""
        return self._request("DELETE", self._base_uri + uri, params=params)

    def post(
        self,
        uri: str,
        *,
        params: Optional[Dict[str, Optional[str]]] = None,
        data: Optional[Union[Dict[str, Any], Iterable[Any]]] = None
    ) -> Awaitable[Tuple[Any, HttpResponse]]:
        """Perform a POST request."""
        return self._request("POST", self._base_uri + uri, params=params, data=data)

    def put(
        self,
        uri: str,
        *,
        params: Optional[Dict[str, Optional[str]]] = None,
        data: Optional[Union[Dict[str, Any], Iterable[Any]]] = None
    ) -> Awaitable[Tuple[Any, HttpResponse]]:
        """Perform a PUT request."""
        return self._request("PUT", self._base_uri + uri, params=params, data=data)

    def patch(
        self,
        uri: str,
        *,
        params: Optional[Dict[str, Optional[str]]] = None,
        data: Optional[Union[Dict[str, Any], Iterable[Any]]] = None
    ) -> Awaitable[Tuple[Any, HttpResponse]]:
        """Perform a PATCH request."""
        return self._request("PATCH", self._base_uri + uri, params=params, data=data)

    @property
    def base_uri(self) -> str:
        return self._base_uri


def _expand_uri_params(
    uri: str, params: Optional[Dict[str, Optional[str]]]
) -> Tuple[str, Optional[Dict[str, str]]]:
    """Expand any params in uri with a url-encoded version of the corresponding value in ``params``.

    Any matched params will be removed from params. Any unmatched params will be left
    as-is in the uri.

    For example, uri can be "/tags/{path}" and params can be {"path": "foo/bar"},
    resulting in the uri "/tags/foo%2Fbar" and params={}.
    """
    if params is None:
        return uri, params
    params2 = {k: v for k, v in params.items() if v is not None}
    if "{" not in uri:
        return uri, params2
    for param, value in params.items():
        param_match = "{" + param + "}"
        if param_match in uri:
            uri = uri.replace(param_match, urllib.parse.quote(value or "", safe=""))
            del params2[param]
    return uri, params2


def _handle_response(response: HttpResponse, method: str, uri: str) -> Any:
    try:
        data = response.json() if len(response.text) > 0 else None
        non_json_error = None
    except json.decoder.JSONDecodeError as ex:
        # For error statuses (e.g. 403), if the body isn't JSON, raise an ApiException
        # with the body text as a message
        data = None
        non_json_error = response.text

        # But if this was supposed to be a non-error, raise the decode error
        if 200 <= response.status_code < 300:
            # TODO: This works around Bug 369006 where SystemLink Cloud returns non-JSON
            # text for some APIs (but only those for which no response is necessary)
            if response.status_code == 200 and response.text in ("Success", "OK"):
                data = None
            elif response.status_code == 201 and response.text.startswith("Created"):
                data = None
            else:
                raise json.decoder.JSONDecodeError(
                    "Error from <{} {}>: {}\n\nResponse text:\n  {}".format(
                        method, uri, ex.args[0], response.text.replace("\n", "\n  ")
                    ),
                    ex.doc,
                    ex.pos,
                ) from None

    if not 200 <= response.status_code < 300:
        msg = "Server responded with <{} {}> when calling {} ({})".format(
            response.status_code,
            getattr(response, "reason_phrase", None) or getattr(response, "reason"),
            method,
            uri,
        )
        if non_json_error:
            msg += ":\n\n" + non_json_error

        if data:
            err_dict = typing.cast(Dict[str, Any], data).get("error", {})
            err_obj = core.ApiError.parse_obj(err_dict) if err_dict else None
        else:
            err_obj = None

        raise core.ApiException(
            msg, error=err_obj, http_status_code=response.status_code
        )

    return data



# -*- coding: utf-8 -*-

"""Implementation of TagSubscription."""

import abc
import contextlib
import datetime
import weakref
from types import TracebackType
from typing import Iterable, List, Optional, Type

import events
from nisystemlink.clients import core, tag as tbase
from nisystemlink.clients.tag._core._manual_reset_timer import ManualResetTimer


class TagSubscription(events.Events, abc.ABC):
    """Represents a subscription for changes to one or more tags' values.

    Call :meth:`close()` to stop receiving events.

    Note that :class:`TagSubscription` objects support using the ``with`` statement (or
    the ``async with`` statement), to :meth:`close()` the subsription automatically on
    exit.

    Attributes:
        tag_changed: An event that is triggered when one of the subscription's tag
            changes. The callback will receive a :class:`TagData` parameter and an
            :class:`Optional` [:class:`TagValueReader`] parameter.

            Example::

                def my_callback(tag: TagData, reader: Optional[TagValueReader]):
                    print("{} changed".format(tag.path))
                    if reader is None:
                        print(" - unknown data type")
                    else:
                        value = reader.read()
                        assert value is not None
                        print(" - new value: {}".format(value.value))

                subscription.tag_changed += my_callback
    """

    __events__ = ["tag_changed"]
    # Under certain circumstances, mypy complains about the event not having a type hint
    # unless we specify it explicitly. (But we also need to delete the attribute so that
    # Events.__getattr__ can do its magic.)
    tag_changed = None  # type: events._EventSlot
    del tag_changed

    _HEARTBEAT_INTERVAL_MILLISECONDS = 30000.0
    """Send a heartbeat every 30 seconds based on a server-side expiration of 60 seconds."""

    def __init__(
        self, paths: Iterable[str], heartbeat_timer: Optional[ManualResetTimer]
    ) -> None:
        """Initialize the instance.

        Derived types must call :meth:`_initialize()` or :meth:`_initialize_async()`
        after construction.

        Args:
            paths: The tag path queries to include in the subscription.
            heartbeat_timer: A timer for sending a heartbeat to keep the subscription
                alive for testing purposes, or None to use a default timer.

        Raises:
            ValueError: if ``paths`` is None.
        """
        if paths is None:
            raise ValueError("paths cannot be None")

        super().__init__()
        self._paths = list(paths)
        if heartbeat_timer is not None:
            self._heartbeat_timer = heartbeat_timer
        else:
            self._heartbeat_timer = ManualResetTimer(
                datetime.timedelta(milliseconds=self._HEARTBEAT_INTERVAL_MILLISECONDS)
            )
        self._exit_stack = contextlib.ExitStack()
        self._exit_stack.enter_context(self._heartbeat_timer)

        callback_ref = weakref.WeakMethod(self._heartbeat_timer_elapsed)  # type: ignore

        def callback() -> None:
            actual_callback = callback_ref()  # type: ignore
            if actual_callback:
                actual_callback()

        self._heartbeat_timer_handler = callback
        self._heartbeat_timer.elapsed += self._heartbeat_timer_handler
        self._closed = False

    def __del__(self) -> None:
        self._exit_stack.close()

    def _initialize(self) -> None:
        """Create and initialize the subscription.

        Derived types must call this method or :meth:`_initialize_async()` after
        construction to create and keep the subscription alive.

        Raises:
            ApiException: if the API call fails.
        """
        self._create_subscription_on_server(self._paths)
        self._heartbeat_timer.start()

    async def _initialize_async(self) -> None:
        """Asynchronously create and initializes the subscription.

        Derived types must call this method or :meth:`_initialize()` after construction
        to create and keep the subscription alive.

        Raises:
            ApiException: if the API call fails.
        """
        await self._create_subscription_on_server_async(self._paths)
        self._heartbeat_timer.start()

    @abc.abstractmethod
    def _create_subscription_on_server(self, paths: List[str]) -> None:
        """Create the subscription on the server.

        Implementations should retrieve and throw away the first set of updates before
        returning.

        Args:
            paths: The tag path queries to include in the subscription.

        Raises:
            ApiException: if the API call fails.
        """
        ...

    @abc.abstractmethod
    async def _create_subscription_on_server_async(self, paths: List[str]) -> None:
        """Asynchronously create the subscription on the server.

        Implementations should retrieve and throw away the first set of updates before
        returning.

        Args:
            paths: The tag path queries to include in the subscription.

        Returns:
            A task representing the asynchronous operation.

        Raises:
            ApiException: if the API call fails.
        """
        ...

    @abc.abstractmethod
    def _send_heartbeat(self) -> None:
        """Send a heartbeat for the subscription to keep it active.

        Raises:
            ApiException: if the API call fails.
        """
        ...

    @abc.abstractmethod
    def _close_internal(self) -> None:
        """Clean up server resources associated with the subscription."""
        ...

    @abc.abstractmethod
    async def _close_internal_async(self) -> None:
        """Asynchronously clean up server resources associated with the subscription.

        Returns:
            A task representing the asynchronous operation.
        """
        ...

    def close(self) -> None:
        """Close server resources associated with the subscription.

        Further tag writes will not trigger new events.
        """
        if self._closed:
            return

        self._close_internal()
        self._heartbeat_timer.elapsed -= self._heartbeat_timer_handler
        self._closed = True

    async def close_async(self) -> None:
        """Asynchronously close server resources associated with the subscription.

        Further tag writes will not trigger new events.

        Returns:
            A task representing the asynchronous operation.
        """
        if self._closed:
            return

        await self._close_internal_async()
        self._heartbeat_timer.elapsed -= self._heartbeat_timer_handler
        self._closed = True

    def __enter__(self) -> "TagSubscription":
        return self

    async def __aenter__(self) -> "TagSubscription":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> bool:
        """Close server resources associated with the subscription.

        Further tag writes will not trigger new events.
        """
        suppress = False
        try:
            self.close()
        finally:
            suppress = self._exit_stack.__exit__(exc_type, exc_val, exc_tb)
        return suppress

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> bool:
        """Asynchronously close server resources associated with the subscription.

        Further tag writes will not trigger new events.
        """
        suppress = False
        try:
            await self.close_async()
        finally:
            suppress = self._exit_stack.__exit__(exc_type, exc_val, exc_tb)
        return suppress

    def _on_tag_changed(
        self, tag: tbase.TagData, value: Optional[tbase.TagValueReader]
    ) -> None:
        """Raise the :attr:`tag_changed` event.

        Args:
            tag: The tag that was changed.
            value: The new value and any associated information, or None if the tag has
                an unknown data type.
        """
        self.tag_changed(tag, value)

    def _heartbeat_timer_elapsed(self) -> None:
        try:
            self._send_heartbeat()
        except core.ApiException:
            try:
                self._create_subscription_on_server(self._paths)
            except core.ApiException:
                # Ignore, we'll try again later
                pass

        self._heartbeat_timer.start()



# -*- coding: utf-8 -*-

"""Implementation of ITimeStamper."""

import abc
import datetime


class ITimeStamper(abc.ABC):
    """Represents an object for time-stamping objects or actions such that no two
    timestamps returned by the instance are within the same microsecond.
    """

    @property
    @abc.abstractmethod
    def timestamp(self) -> datetime.datetime:  # noqa: D401
        """A unique UTC ``datetime.datetime`` that is close to the current date and time."""
        ...



from enum import Enum
from typing import List, Optional, Union

from nisystemlink.clients.core._uplink._json_model import JsonModel


class ConditionType(Enum):
    """Conditions are either numeric or string type."""

    NUMERIC = "NUMERIC"
    """Numeric condition."""

    STRING = "STRING"
    """String condition."""


class ConditionRange(JsonModel):
    """Specifies the range of values that the condition must cover."""

    min: Optional[float] = None
    """Minimum value of the condition range."""

    max: Optional[float] = None
    """Maximum value of the condition range."""

    step: Optional[float] = None
    """Step value of the condition range."""


class ConditionValueBase(JsonModel):
    """The base type for conditions that can be represented in several styles."""

    condition_type: ConditionType
    """Type of the Condition."""


class NumericConditionValue(ConditionValueBase):
    """A numeric condition.

    Numeric conditions can contain a combination of ranges and discrete lists.
    """

    range: Optional[List[ConditionRange]] = None
    """List of condition range values."""

    discrete: Optional[List[float]] = None
    """List of condition discrete values."""

    unit: Optional[str] = None
    """Unit of the condition."""


class StringConditionValue(ConditionValueBase):
    """A string condition.

    String conditions may only contain discrete lists of values.
    """

    discrete: Optional[List[str]] = None
    """List of condition discrete values."""


class Condition(JsonModel):
    """A single condition."""

    name: Optional[str] = None
    """Name of the condition."""

    value: Optional[Union[NumericConditionValue, StringConditionValue]] = None
    """Value of the condition."""



from typing import Dict, List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel


class TableMetadataModification(JsonModel):
    """Contains the metadata properties to modify. Values not included in the
    request or included with a ``None`` value will remain unchanged.
    """

    id: str
    """The ID of the table to modify."""

    metadata_revision: Optional[int] = None
    """When specified, this is an integer that must match the last known
    revision number of the table, incremented by one. If it doesn't match the
    current ``metadataRevision`` incremented by one at the time of execution, the
    modify request will be rejected with a conflict error. This is used to
    ensure that changes to this table's metadata are based on a known, previous
    state."""

    name: Optional[str] = None
    """The new name of the table."""

    workspace: Optional[str] = None
    """The new workspace for the table. Changing the workspace requires
    permission to delete the table in its current workspace and permission to
    create the table in its new workspace."""

    properties: Optional[Dict[str, Optional[str]]] = None
    """The properties to modify. A map of key value properties containing the
    metadata to be added or modified. Setting a property value to ``None`` will
    delete the property. Existing properties not included in the map are
    unaffected unless replace is true in the top-level request object."""


class ModifyTablesRequest(JsonModel):
    """Contains one or more table modifications to apply."""

    tables: List[TableMetadataModification]
    """The table modifications to apply. Each table may only appear once in the list."""

    replace: Optional[bool] = None
    """When true, existing properties are replaced instead of merged."""



from nisystemlink.clients.core._api_info import Operation
from nisystemlink.clients.core._uplink._json_model import JsonModel


class V2Operations(JsonModel):
    """The operations available in the routes provided by the v2 HTTP API."""

    get_products: Operation
    """The ability to get a list of products."""

    query_products: Operation
    """The ability to query products based on their properties."""

    create_products: Operation
    """The ability to create one or more products."""

    update_products: Operation
    """The ability to update the properties of one or more products."""

    delete_products: Operation
    """The ability to delete a single products."""

    delete_many_products: Operation
    """The ability to delete a list of products."""

    get_results: Operation
    """The ability to get a list of results."""

    get_results_property_keys: Operation
    """The ability to get custom property keys."""

    query_results: Operation
    """"The ability to to query results based on their properties."""

    create_results: Operation
    """The ability to create results."""

    update_results: Operation
    """The ability to update results."""

    delete_result: Operation
    """The ability to delete a single results."""

    delete_many_results: Operation
    """The ability to delete multiple results."""

    get_steps: Operation
    """The ability to get a list of steps."""

    query_steps: Operation
    """The ability to query steps based on their properties."""

    create_steps: Operation
    """The ability to create steps."""

    update_steps: Operation
    """The ability to update steps."""

    delete_step: Operation
    """The ability to delete a single step."""

    delete_many_steps: Operation
    """The ability to delete multiple steps."""

    query_paths: Operation
    """The ability to query step paths."""


class ApiInfo(JsonModel):
    """Information about the available API operations."""

    operations: V2Operations



# -*- coding: utf-8 -*-

"""Implementation of BufferedTagWriter."""

import abc
import datetime
import sys
import threading
from types import TracebackType
from typing import Any, Callable, Optional, Type

from nisystemlink.clients import core, tag as tbase
from nisystemlink.clients.tag._core._itime_stamper import ITimeStamper
from nisystemlink.clients.tag._core._manual_reset_timer import ManualResetTimer


class BufferedTagWriter(tbase.ITagWriter):
    """Represents an :class:`ITagWriter` that buffers tag writes instead of sending them immediately.

    Writes that utilize automatic timestamps are based on the system time when buffered.
    Implementations may provide automatic sending of buffered writes based on different
    conditions. Unsent writes are discarded when the instance is deleted.

    Note that :class:`BufferedTagWriter` objects support using the ``with`` statement
    (or the ``async with`` statement), to automatically :meth:`send
    <send_buffered_writes>` any remaining buffered writes on exit.
    """

    def __init__(
        self, stamper: ITimeStamper, buffer_size: int, flush_timer: ManualResetTimer
    ) -> None:
        """Initialize the writer.

        Args:
            stamper: An object for time-stamping tag writes.
            buffer_size: The maximum number of tag writes to buffer before automatically
                sending them to the server.
            flush_timer: A timer that, once started, elapses whenever buffered writes
                should be sent automatically. Does not have to be a configured timer.
        """
        self._lock = threading.Lock()
        self._buffer_limit = buffer_size
        self._flush_timer = flush_timer
        self._stamper = stamper

        self._closed = False
        self._num_buffered = 0
        self._send_error = None  # type: Optional[core.ApiException]
        self._timer_generation = 0
        self._timer_handler = None  # type: Optional[Callable[[], None]]

    @abc.abstractmethod
    def _buffer_value(self, path: str, value: Any) -> None:
        """Add a value to the buffer.

        Args:
            path: The tag path being written.
            value: The value being written.
        """
        ...

    @abc.abstractmethod
    def _clear_buffer(self) -> None:
        """Clear the buffer of writes."""
        ...

    @abc.abstractmethod
    def _copy_buffer(self) -> Any:
        """Return the contents of the buffer and clears or replaces the buffer used for future writes.

        Returns:
            The buffered data.
        """
        ...

    @abc.abstractmethod
    def _create_item(
        self,
        path: str,
        data_type: tbase.DataType,
        value: str,
        timestamp: Optional[datetime.datetime] = None,
    ) -> Any:
        """Return an item that can be placed into the buffer.

        Args:
            path: The path of the tag to write.
            data_type: The data type of the value to write.
            value: The tag value to write, serialized as a string.
            timestamp: The timestamp represented as a nullable ``datetime.datetime``.

        Returns:
            The created item.

        Raises:
            ValueError: if ``data_type`` is not supported for writing.
        """
        ...

    @abc.abstractmethod
    def _send_writes(self, updates: Any) -> None:
        """Send the writes stored in ``updates`` to the server.

        Args:
            updates: The writes to send.

        Raises:
            ApiException: if the API call fails.
        """
        ...

    @abc.abstractmethod
    async def _send_writes_async(self, updates: Any) -> None:
        """Asynchronously send the writes stored in ``updates`` to the server.

        Args:
            updates: The writes to send.

        Returns:
            A task representing the asynchronous operation.

        Raises:
            ApiException: if the API call fails.
        """
        ...

    def clear_buffered_writes(self) -> None:
        """Clear any pending writes from :meth:`write()`.

        Raises:
            ReferenceError: if the writer has been closed.
        """
        if self._closed:
            raise ReferenceError("BufferedTagWriter")

        with self._lock:
            self._stop_timer_while_locked()
            self._clear_buffer()
            self._num_buffered = 0

    def send_buffered_writes(self) -> None:
        """Write all of the pending writes from :meth:`write()` to the server.

        Does nothing if there are no pending writes.

        Raises:
            ReferenceError: if the writer has been closed.
            ApiException: if the API call fails.
        """
        if self._closed:
            raise ReferenceError("BufferedTagWriter")

        with self._lock:
            updates = self._retrieve_buffered_values_while_locked()

        if updates is not None:
            self._send_writes(updates)

    async def send_buffered_writes_async(self) -> None:
        """Asynchronously write all of the pending writes from :meth:`write()` to the server.

        Does nothing if there are no pending writes.

        Raises:
            ReferenceError: if the writer has been closed.
            ApiException: if the API call fails.
        """
        if self._closed:
            raise ReferenceError("BufferedTagWriter")

        with self._lock:
            updates = self._retrieve_buffered_values_while_locked()

        if updates is not None:
            await self._send_writes_async(updates)

    def __enter__(self) -> "BufferedTagWriter":
        if self._closed:
            raise ReferenceError("BufferedTagWriter")

        self._flush_timer.__enter__()
        return self

    async def __aenter__(self) -> "BufferedTagWriter":
        if self._closed:
            raise ReferenceError("BufferedTagWriter")

        await self._flush_timer.__aenter__()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> bool:
        if self._closed:
            return False

        self.send_buffered_writes()
        self._closed = True

        suppress = self._flush_timer.__exit__(exc_type, exc_val, exc_tb)
        return suppress

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> bool:
        if self._closed:
            return False

        await self.send_buffered_writes_async()
        self._closed = True

        suppress = await self._flush_timer.__aexit__(exc_type, exc_val, exc_tb)
        return suppress

    def _write(
        self,
        path: str,
        data_type: tbase.DataType,
        value: str,
        timestamp: Optional[datetime.datetime] = None,
    ) -> None:
        """Write a tag's value that's been serialized to a string.

        Clients do not typically call this method directly. Use a
        :class:`.TagValueWriter` instead.

        Args:
            path: The path of the tag to write.
            data_type: The data type of the value to write.
            value: The tag value to write, serialized as a string.
            timestamp: A custom timestamp to associate with the value, or None to have
                the server specify the timestamp.

        Raises:
            ValueError: if `path` is empty or invalid.
            ValueError: if `path` or `value` is None.
            ValueError: if `data_type` is invalid.
            ReferenceError: if the writer has been closed.
            ApiException: if the API call fails.
        """
        timestamped_value = self._prepare_write(path, data_type, value, timestamp)

        pending_error = None
        updates = None
        with self._lock:
            self._buffer_value(path, timestamped_value)
            self._num_buffered += 1

            if self._num_buffered == self._buffer_limit:
                updates = self._retrieve_buffered_values_while_locked()
            elif self._num_buffered == 1:
                self._start_timer_while_locked()

            if self._send_error is not None:
                pending_error = self._send_error
                self._send_error = None

        if updates is not None:
            self._send_writes(updates)

        if pending_error:
            raise pending_error

    async def _write_async(
        self,
        path: str,
        data_type: tbase.DataType,
        value: str,
        timestamp: Optional[datetime.datetime] = None,
    ) -> None:
        """Asynchronously write a tag's value that's been serialized to a string.

        Clients do not typically call this method directly. Use a
        :class:`.TagValueWriter` instead.

        Args:
            path: The path of the tag to write.
            data_type: The data type of the value to write.
            value: The tag value to write, serialized as a string.
            timestamp: A custom timestamp to associate with the value, or None to have
                the server specify the timestamp.

        Returns:
            A task representing the asynchronous operation.

        Raises:
            ValueError: if `path` is empty or invalid.
            ValueError: if `path` or `value` is None.
            ValueError: if `data_type` is invalid.
            ReferenceError: if the writer has been closed.
            ApiException: if the API call fails.
        """
        timestamped_value = self._prepare_write(path, data_type, value, timestamp)

        pending_error = None
        updates = None
        with self._lock:
            self._buffer_value(path, timestamped_value)
            self._num_buffered += 1

            if self._num_buffered == self._buffer_limit:
                updates = self._retrieve_buffered_values_while_locked()
            elif self._num_buffered == 1:
                self._start_timer_while_locked()

            if self._send_error is not None:
                pending_error = self._send_error
                self._send_error = None

        if updates is not None:
            await self._send_writes_async(updates)

        if pending_error:
            raise pending_error

    def _prepare_write(
        self,
        path: str,
        data_type: tbase.DataType,
        value: str,
        timestamp: Optional[datetime.datetime] = None,
    ) -> Any:
        if self._closed:
            raise ReferenceError("BufferedTagWriter")

        if value is None:
            raise ValueError("value is None")

        if data_type == tbase.DataType.UNKNOWN:
            raise ValueError("data_type is UNKNOWN")

        if timestamp is None:
            timestamp = self._stamper.timestamp
        else:
            if timestamp.tzinfo is None and sys.version_info < (3, 6):
                # python 3.5's .astimezone fails if given a naive datetime, so use a
                # roundabout method; python 3.6+ allow naive datetimes and assume a
                # local timezone, so allow the same here
                timestamp = datetime.datetime.fromtimestamp(
                    timestamp.timestamp(), datetime.timezone.utc
                )
            else:
                timestamp = timestamp.astimezone(datetime.timezone.utc)
        return self._create_item(
            tbase.TagPathUtilities.validate(path), data_type, value, timestamp
        )

    def _retrieve_buffered_values_while_locked(self) -> Any:
        """Return the buffered values, if any, and clears the buffer.

        Must hold :attr:`_lock`.

        Returns:
            The buffered values, or None if there aren't any.
        """
        self._stop_timer_while_locked()

        if self._num_buffered == 0:
            return None

        buffer = self._copy_buffer()
        self._num_buffered = 0
        return buffer

    def _start_timer_while_locked(self) -> None:
        """Start the flush timer, if configured.

        Must hold :attr:`_lock`.
        """
        if not self._flush_timer.can_start:
            return

        handler_generation = self._timer_generation
        self._flush_timer.elapsed -= self._timer_handler
        self._timer_handler = lambda: self._timer_expired(handler_generation)
        self._flush_timer.elapsed += self._timer_handler
        self._flush_timer.start()

    def _stop_timer_while_locked(self) -> None:
        """Stop the flush timer, if configured.

        Must hold :attr:`_lock`.
        """
        self._flush_timer.stop()
        self._timer_generation += 1

    def _timer_expired(self, generation: int) -> None:
        if self._closed:
            return

        with self._lock:
            if generation != self._timer_generation:
                # The timer was canceled after we were already queued.
                return

            updates = self._retrieve_buffered_values_while_locked()

        if updates is not None:
            try:
                self._send_writes(updates)
            except core.ApiException as ex:
                with self._lock:
                    self._send_error = ex



# -*- coding: utf-8 -*-

"""Implementation of SerializedTagWithAggregatesReader."""

from typing import Optional

from nisystemlink.clients import tag as tbase
from nisystemlink.clients.tag._core._serialized_tag_with_aggregates import (
    SerializedTagWithAggregates,
)
from typing_extensions import final


@final
class SerializedTagWithAggregatesReader(tbase.ITagReader):
    """Represents an :class:`.ITagReader` wrapping a single :class:`SerializedTagWithAggregates`."""

    def __init_subclass__(cls) -> None:
        raise TypeError(
            "type 'SerializedTagWithAggregatesReader' is not an acceptable base type"
        )

    def __init__(self, value: SerializedTagWithAggregates) -> None:
        """Initialize a reader instance for the given value.

        Args:
            value: The value to wrap in a reader.

        Raises:
            ValueError: if ``value`` is None.
        """
        if value is None:
            raise ValueError("value cannot be None")
        self._value = value

    def _read(
        self, path: str, include_timestamp: bool, include_aggregates: bool
    ) -> Optional[SerializedTagWithAggregates]:
        """Retrieve the current value of the tag with the given ``path`` from the server.

        Optionally retrieves the aggregate values as well. The tag must exist. Clients
        do not typically call this method directly. Use a :class:`.TagValueReader`
        instead.

        Args:
            path: The path of the tag to read.
            include_timestamp: True to include the timestamp associated with the value
                in the result.
            include_aggregates: True to include the tag's aggregate values in the result
                if the tag is set to :attr:`TagData.collect_aggregates`.

        Returns:
            The value serialized as a string, or None if the tag exists but doesn't have
            a value.
        """
        if self._value.path == path:
            return self._value
        else:
            return None

    async def _read_async(
        self, path: str, include_timestamp: bool, include_aggregates: bool
    ) -> Optional[SerializedTagWithAggregates]:
        """Asynchronously retrieve the current value of the tag with the given ``path`` from the server.

        Optionally retrieves the aggregate values as well. The tag must exist. Clients
        do not typically call this method directly. Use a :class:`.TagValueReader`
        instead.

        Args:
            path: The path of the tag to read.
            include_timestamp: True to include the timestamp associated with the value
                in the result.
            include_aggregates: True to include the tag's aggregate values in the result
                if the tag is set to :attr:`TagData.collect_aggregates`.

        Returns:
            The value serialized as a string, or None if the tag exists but doesn't have
            a value.
        """
        if self._value.path == path:
            return self._value
        else:
            return None



from typing import Optional

from ._json_model import JsonModel


class WithPaging(JsonModel):
    continuation_token: Optional[str] = None
    """A token which allows the user to resume a query at the next item in the matching results.

    When querying, a token will be returned if a query may be
    continued. To obtain the next page of results, pass the token to the service
    on a subsequent request. The service will respond with a new continuation
    token. To paginate results, continue sending requests with the newest
    continuation token provided by the service, until this value is null.
    """



from typing import Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel


class ColumnOrderBy(JsonModel):
    """Specifies a column to order by and the ordering direction."""

    column: str
    """The name of the column to order by."""

    descending: Optional[bool] = None
    """Whether the ordering should be in descending order."""
